theory Quasi_Reductivity
imports
  "../Rewriting/Context_Sensitive"
  Conditional_Rewriting
begin

locale replacement_map =
  fixes \<mu> :: "'f mu"
begin

inductive_set
  supt_mu :: "('f, 'v) term rel" ("{\<rhd>\<^sub>\<mu>}")
where
  arg [simp, intro]: "\<And>f n i ss\<^sub>1 s ss\<^sub>2. \<lbrakk>i \<in> \<mu> (f, n); i = length ss\<^sub>1; n = i + length ss\<^sub>2 + 1\<rbrakk> \<Longrightarrow>
    (Fun f (ss\<^sub>1 @ s # ss\<^sub>2), s) \<in> {\<rhd>\<^sub>\<mu>}"
| trans: "(s, t) \<in> {\<rhd>\<^sub>\<mu>} \<Longrightarrow> (t, u) \<in> {\<rhd>\<^sub>\<mu>} \<Longrightarrow> (s, u) \<in> {\<rhd>\<^sub>\<mu>}"

abbreviation "supt_mu_pred s t \<equiv> (s, t) \<in> {\<rhd>\<^sub>\<mu>}"

notation
  supt_mu_pred ("(_/ \<rhd>\<^sub>\<mu> _)" [56, 56] 55)

inductive_set mu_ctxts
where
  Hole [simp, intro]: "\<box> \<in> mu_ctxts"
| More [intro]: "\<And>f n i ss\<^sub>1 C ss\<^sub>2. \<lbrakk>i \<in> \<mu> (f, n); i = length ss\<^sub>1; n = i + length ss\<^sub>2 + 1; C \<in> mu_ctxts\<rbrakk> \<Longrightarrow>
  More f ss\<^sub>1 C ss\<^sub>2 \<in> mu_ctxts"

inductive_set murstep for R
where
  murstep: "C \<in> mu_ctxts \<Longrightarrow> (l, r) \<in> R \<Longrightarrow> s = C\<langle>l \<cdot> \<sigma>\<rangle> \<Longrightarrow> t = C\<langle>r \<cdot> \<sigma>\<rangle> \<Longrightarrow> (s, t) \<in> murstep R"

lemma mu_ctxts_compose [intro]:
  "C \<in> mu_ctxts \<Longrightarrow> D \<in> mu_ctxts \<Longrightarrow> C \<circ>\<^sub>c D \<in> mu_ctxts"
by (induct C rule: mu_ctxts.induct) auto

lemma murstep_ctxt:
  assumes "C \<in> mu_ctxts" and "(s, t) \<in> murstep R"
  shows "(C\<langle>s\<rangle>, C\<langle>t\<rangle>) \<in> murstep R"
proof -
  obtain D l r \<sigma> where "D \<in> mu_ctxts" and "(l, r) \<in> R" and "s = D\<langle>l \<cdot> \<sigma>\<rangle>" and "t = D\<langle>r \<cdot> \<sigma>\<rangle>"
    using assms(2) by (cases) auto
  with assms show ?thesis
  proof (induct)
    case (More f n i ss\<^sub>1 C ss\<^sub>2)
    then show ?case
      by (intro murstep.murstep [where C = "More f ss\<^sub>1 C ss\<^sub>2 \<circ>\<^sub>c D"]) auto
  qed simp
qed

lemma mu_ctxts_imp_supt_mu:
  "C \<in> mu_ctxts \<Longrightarrow> C \<noteq> \<box> \<Longrightarrow> C\<langle>t\<rangle> \<rhd>\<^sub>\<mu> t"
proof (induct C rule: mu_ctxts.induct)
  case (More f n i ss\<^sub>1 C ss\<^sub>2)
  moreover then have "(More f ss\<^sub>1 C ss\<^sub>2)\<langle>t\<rangle> \<rhd>\<^sub>\<mu> C\<langle>t\<rangle>" by auto
  ultimately show ?case by (cases "C = \<box>") (simp, blast dest: supt_mu.trans)
qed simp

lemma supt_mu_iff:
  "s \<rhd>\<^sub>\<mu> t \<longleftrightarrow> (\<exists>C \<in> mu_ctxts. C \<noteq> \<box> \<and> s = C\<langle>t\<rangle>)" (is "_ = ?R")
proof
  assume "s \<rhd>\<^sub>\<mu> t"
  then show ?R
  proof (induct)
    case (arg f n i ss\<^sub>1 s ss\<^sub>2)
    then show ?case by (intro bexI [where x = "More f ss\<^sub>1 \<box> ss\<^sub>2"]) auto
  next
    case (trans s t u)
    then obtain C and D where "C \<in> mu_ctxts" and "C \<noteq> \<box>" and "s = C\<langle>t\<rangle>"
      and "D \<in> mu_ctxts" and "D \<noteq> \<box>" and "t = D\<langle>u\<rangle>" by blast
    then show ?case
      by (intro bexI [where x = "C \<circ>\<^sub>c D"]) (auto elim: ctxt_compose.elims)
  qed
qed (auto intro: mu_ctxts_imp_supt_mu)

lemma supt_mu_murstep_commute:
  "{\<rhd>\<^sub>\<mu>} O murstep R \<subseteq> murstep R O {\<rhd>\<^sub>\<mu>}"
proof
  fix s u :: "('f, 'v) term" assume "(s, u) \<in> {\<rhd>\<^sub>\<mu>} O murstep R"
  then obtain t where "s \<rhd>\<^sub>\<mu> t" and "(t, u) \<in> murstep R" by blast
  moreover then obtain C where "C \<noteq> \<box>" and "C \<in> mu_ctxts" and s: "s = C\<langle>t\<rangle>" by (auto simp: supt_mu_iff)
  ultimately have "(C\<langle>t\<rangle>, C\<langle>u\<rangle>) \<in> murstep R" and "C\<langle>u\<rangle> \<rhd>\<^sub>\<mu> u"
    by (auto intro: murstep_ctxt simp: supt_mu_iff)
  then show "(s, u) \<in> murstep R O {\<rhd>\<^sub>\<mu>}" unfolding s by blast
qed

context
  fixes U :: "('f, 'v) ctrs \<Rightarrow> ('f, 'v) trs"
  assumes simpcomp: "cstep R \<subseteq> (murstep (U R))\<^sup>+"
begin

lemma
  fixes R :: "('f, 'v) ctrs"
  assumes "funas_ctrs R \<subseteq> F"
    and "SN_on (murstep (U R)) {t::('f, 'v) term. funas_term t \<subseteq> F}" (is "SN_on _ ?T")
  shows "quasi_decreasing R"
proof -
  def less (infix "\<succ>" 50) \<equiv> "Restr ((murstep (U R) \<union> {\<rhd>\<^sub>\<mu>})\<^sup>+) ?T"
  have "trans less" by (auto simp: trans_def less_def)
  moreover have "SN less" sorry
  moreover have "cstep R \<subseteq> less"
    using simpcomp [of R, THEN subsetD, THEN trancl_mono] apply (auto simp: less_def)
    sorry
  moreover
  have "quasi_decreasing_order R less" sorry

(*
  (trans S \<and> SN S \<and> 
  cstep R \<subseteq> S \<and> {\<rhd>} \<subseteq> S \<and> 
  (\<forall> l r cs (\<sigma> :: ('f, 'v) subst). ((l, r), cs) \<in> R \<longrightarrow> 
    (\<forall> i. i < length cs \<longrightarrow> (\<forall>j < i. (fst (cs ! j) \<cdot> \<sigma>, snd (cs ! j) \<cdot> \<sigma>) \<in> (cstep R)\<^sup>* ) \<longrightarrow>
      (l \<cdot> \<sigma>, fst (cs ! i) \<cdot> \<sigma>) \<in> S)))"
*)
  then show ?thesis by (auto simp: quasi_decreasing_def)
qed

end

end
