theory Bla
  imports "HOL-Library.Multiset"
begin

\<comment> \<open>All possible ways to insert a single element into a list\<close>
fun inserts :: "'a \<Rightarrow> 'a list \<Rightarrow> 'a list list"
  where
    "inserts x [] = [[x]]"
  | "inserts x (y # ys) = (x # y # ys) # map ((#) y) (inserts x ys)"

\<comment> \<open>\<open>ins xs y zs\<close> holds if \<open>xs\<close> results from inserting \<open>y\<close> into \<open>zs\<close> at an arbitrary position.\<close>
inductive ins :: "'a list \<Rightarrow> 'a \<Rightarrow> 'a list \<Rightarrow> bool"
  where
    "x = y \<Longrightarrow> ins (y # ys) x ys"
  | "y = z \<Longrightarrow> ins ys x zs \<Longrightarrow> ins (y # ys) x (z # zs)"

lemma ins_Cons [simp]:
  "ins (y # ys) x (y # zs) \<longleftrightarrow> ins ys x zs"
  by (auto elim: ins.cases intro: ins.intros)

lemma in_set_inserts_iff:
  "ys \<in> set (inserts x zs) \<longleftrightarrow> ins ys x zs"
  by (induct zs arbitrary: ys) (auto intro: ins.intros elim: ins.cases)

lemma ins_mset [simp]:
  assumes "ins ys x zs"
  shows "mset ys = add_mset x (mset zs)"
  using assms by (induct) auto

lemma ins_remove1I:
  assumes "x \<in> set xs"
  shows "ins xs x (remove1 x xs)"
  using assms by (induct xs) (auto intro: ins.intros)

lemma mset_insE:
  assumes "mset xs = add_mset y (mset ys)"
  obtains zs where "mset zs = mset ys" and "ins xs y zs"
  using assms
proof (induct xs arbitrary: ys thesis)
  case (Cons x xs)
  show ?case
  proof (cases "x = y")
    assume "x = y"
    then show ?thesis using Cons by (auto intro: ins.intros)
  next
    case False
    with Cons.prems have *: "mset (x # remove1 y xs) = mset ys" and "y \<in> set xs"
      by (auto simp: diff_union_swap) (meson in_multiset_in_set insert_noteq_member)
    then show ?thesis
      by (intro Cons.prems(1) [OF *]) (auto intro: ins_remove1I)
  qed
qed simp

fun perms :: "'a list \<Rightarrow> 'a list list"
  where
    "perms [] = [[]]"
  | "perms (x#xs) = concat (map (inserts x) (perms xs))"

lemma perms [simp]:
  "set (perms xs) = {ys. mset ys = mset xs}"
  by (induct xs) (fastforce simp: in_set_inserts_iff elim: mset_insE)+

end
