(*
Author:  Bertram Felgenhauer <bertram.felgenhauer@uibk.ac.at> (2015)
Author:  Makarius Wenzel <makarius@sketis.net> (2013)
Author:  René Thiemann <rene.thiemann@uibk.ac.at> (2013-2015)
License: LGPL (see file COPYING.LESSER)
*)
theory Container_Setup
imports 
  "Abstract-Rewriting.SN_Order_Carrier"
  Simplex.Simplex
  Containers.Containers
(*  Algebraic_Numbers.Complex_Algebraic_Numbers *)
  Jordan_Normal_Form.Matrix_IArray_Impl
  Ord.Linear_Polynomial
  TRS.Forbidden_Patterns
  TA.Tree_Automata_Det_Container
  LTS.Cooperation_Program
  Ord.Integer_Arithmetic
  Sem_Lab.Labelings_Impl
  TRS.Ground_Context_Impl
begin

  
(* LTS *)
derive compare IA.sig IA.ty
derive ccompare sharp formula trans_var transition_rule IA.sig IA.ty IA.val
derive (rbt) mapping_impl sharp formula trans_var transition_rule IA.sig IA.ty IA.val
derive (eq) ceq sharp formula trans_var transition_rule IA.sig IA.ty IA.val
derive (rbt) set_impl sharp formula trans_var transition_rule IA.sig IA.ty
derive (dlist) set_impl IA.val
derive (no) cenum sharp formula trans_var transition_rule IA.sig IA.ty IA.val

(* term, etc *)
derive (eq) ceq 
  lab "term" mctxt ctxt gctxt location l_poly ta_rule rat arctic arctic_delta
derive compare (* for these types there is already an order defined, so no compare_order poss. *) 
  mctxt arctic arctic_delta
derive compare_order
  ctxt location l_poly
derive (linorder) compare rat real
derive (compare) ccompare real
derive (rbt) set_impl real
derive (eq) ceq real
derive (compare) ccompare 
  lab "term" mctxt ctxt gctxt location l_poly rat arctic arctic_delta
(* Note that the powerset construction uses sets to represent states. Since we cannot instantiate
 * compare for sets, it's necessary to derive a ccompare instance for ta_rule that does not
 * require compare. *)
derive ccompare 
  ta_rule
derive (rbt) set_impl 
  lab "term" mctxt ctxt gctxt location l_poly ta_rule rat arctic arctic_delta
derive (no) cenum 
  lab "term" mctxt ctxt gctxt location l_poly ta_rule rat real arctic arctic_delta
derive (rbt) mapping_impl
  lab

instantiation lab :: (type,type)card_UNIV
begin
definition "card_UNIV = Phantom(('a,'b)lab) 0"
definition "finite_UNIV = Phantom(('a,'b)lab) False"
instance
  by (intro_classes, auto simp: finite_UNIV_lab_def card_UNIV_lab_def infinite_lab)
end


instantiation lab :: (compare,compare)cproper_interval
begin
definition "cproper_interval = (\<lambda> ( _ :: ('a,'b)lab option) _ . False)"
instance by (intro_classes, auto simp: infinite_lab)
end


instantiation ta_rule :: (type,type)finite_UNIV
begin
definition "finite_UNIV = Phantom(('a,'b)ta_rule) False"
instance
  by (intro_classes, unfold finite_UNIV_ta_rule_def, simp)
end

instantiation ta_rule :: (ccompare,ccompare)cproper_interval
begin
definition "cproper_interval = (\<lambda> ( _ :: ('a,'b)ta_rule option) _ . False)"
instance by (intro_classes, auto)
end


lemma infinite_term_UNIV[simp, intro]: "infinite (UNIV :: ('f,'v)term set)"
proof -
  fix f :: 'f and v :: 'v
  let ?inj = "\<lambda>n. Fun f (replicate n (Var v))"
  have "inj ?inj" unfolding inj_on_def by auto
  from infinite_super[OF _ range_inj_infinite[OF this]]
  show ?thesis by blast
qed

instantiation "term" :: (type,type)finite_UNIV
begin
definition "finite_UNIV = Phantom(('a,'b)term) False"
instance
  by (intro_classes, unfold finite_UNIV_term_def, simp)
end

instantiation "term" :: (compare,compare)cproper_interval
begin
definition "cproper_interval = (\<lambda> ( _ :: ('a,'b)term option) _ . False)"
instance by (intro_classes, auto)
end

derive (eq) ceq atom QDelta
derive (linorder) compare_order QDelta
derive compare_order atom
derive ccompare atom QDelta
derive (rbt) set_impl atom QDelta

text \<open>code equation to handle @{term Union}\<close>
lemma un_foldr[code_unfold]: "\<Union> (set xs) = foldr (\<union>) xs {}"
  by (induct xs, auto)

end
