theory XTC_Problem
imports
  CTRS.Conditional_Rewriting_Impl
  Ord.Integer_Arithmetic
  LTS.LTS
  Check_Common
  TRS.Context_Sensitive_Impl
begin

interpretation IA: lts
  where I = IA.I
    and type_of_fun = IA.type_of_fun
    and Values_of_type = IA.Values_of_type
    and Bool_types = "{IA.BoolT}"
    and to_bool = IA.to_bool
    and type_fixer = "TYPE(_)".

datatype (dead 'f, dead 'v) strategy =
  No_Strategy
| Innermost
| Innermost_Q "('f, 'v) term list"

datatype start_term = Full | Constructor_Based

primrec strategy_to_Q :: "('f, 'v) strategy \<Rightarrow> ('f, 'v) rules \<Rightarrow> ('f, 'v) term list" where
  "strategy_to_Q No_Strategy _ = []"
| "strategy_to_Q Innermost R = remdups (map fst R)"
| "strategy_to_Q (Innermost_Q Q) _ = Q"

datatype (dead 'f, dead 'v) fp_strategy =
  Outermost
| Context_Sensitive "'f muL"
| Forbidden_Patterns "('f, 'v) forb_pattern list"

definition "fp_strategy_to_fp_impl strat R \<equiv> case strat of
    Outermost \<Rightarrow> o_to_fp_impl (map fst R)
  | Context_Sensitive \<mu> \<Rightarrow> mu_to_fp_impl \<mu>
  | Forbidden_Patterns fp \<Rightarrow> fp"

definition "fp_strategy_to_string strat \<equiv> case strat of
    Outermost \<Rightarrow> ''outermost''
  | Context_Sensitive _ \<Rightarrow> ''context-sensitive''
  | Forbidden_Patterns _ \<Rightarrow> ''forbidden-pattern''"

datatype (dead 'f, dead 'v) input =
  DP_input "bool" "('f, 'v) rules" "('f, 'v) strategy" "('f, 'v) rules"
| Inn_TRS_input "('f, 'v) strategy" "('f, 'v) rules" "('f, 'v) rules" "start_term"
| CPX_input  "('f, 'v) strategy" "('f, 'v) rules" "('f, 'v) rules" "('f,'v) complexity_measure" complexity_class (* TODO: improve CPF and remove*)
| COMP_input "('f, 'v) equation list" "('f, 'v) rules"
| OCOMP_input "('f, 'v) equation list" "('f, 'v) equation list" "('f, 'v) rules" "'f reduction_order_input"
| EQ_input "('f, 'v) equation list" "('f, 'v) equation_literal"
| FP_TRS_input "('f, 'v) fp_strategy" "('f, 'v) rules"
| CTRS_input "('f, 'v) crules"
| TA_input "(string,'f)tree_automaton" "('f,'v)rules"
| AC_input "('f,'v) rules" "'f list" "'f list"
| LTS_input "(IA.sig, 'v, IA.ty, string, string) lts_impl"
| LTS_safety_input "(IA.sig, 'v, IA.ty, string, string) lts_impl" "string list"
| Unknown_input unknown_info

definition "input_to_string input \<equiv> case input of
  DP_input _ _ _ _ \<Rightarrow> STR ''DP input''
| Inn_TRS_input _ _ _ _ \<Rightarrow> STR ''TRS input''
| CPX_input _ _ _ _ _ \<Rightarrow> STR ''Complexity input''
| COMP_input _ _ \<Rightarrow> STR ''Completion input''
| OCOMP_input _ _ _ _ \<Rightarrow> STR ''Ordered Completion input''
| EQ_input _ _ \<Rightarrow> STR ''Equational input''
| FP_TRS_input _ _ \<Rightarrow> STR ''FP-TRS input''
| CTRS_input _ \<Rightarrow> STR ''CTRS input''
| TA_input _ _ \<Rightarrow> STR ''TA input''
| AC_input _ _ _ \<Rightarrow> STR ''AC-TRS input''
| LTS_input _ \<Rightarrow> STR ''LTS input''
| LTS_safety_input _ _ \<Rightarrow> STR ''LTS safety input''
| Unknown_input _ \<Rightarrow> STR ''Unknown input''"

type_synonym ('f, 'v) inn_fp_strategy = "('f, 'v) strategy + ('f, 'v) fp_strategy"



datatype (dead 'f, dead 'v) claim =
 Yes
| No
| Terminating
| Upperbound nat
| Nonterminating
| Confluent
| Nonconfluent
| Completed (* it should take "('f,'v) trs" as the output complete TRS *)
| Ordered_Completed (* it should take "('f,'v) trs" as the output ground complete system *)
| Anything (* claims whatever a given proof proves *)

definition "claim_to_string claim \<equiv> case claim of
  Terminating \<Rightarrow> STR ''termination''
| Upperbound n \<Rightarrow> STR ''upperbound complexity''
| Nonterminating \<Rightarrow> STR ''nontermination''
| Confluent \<Rightarrow> STR ''confluence''
| Nonconfluent \<Rightarrow> STR ''nonconfluence''
| Completed \<Rightarrow> STR ''completion''
| Ordered_Completed \<Rightarrow> STR ''ordered completion''
| Yes \<Rightarrow> STR ''yes''
| No \<Rightarrow> STR ''no''
| Anything \<Rightarrow> STR ''anything''"


definition "translate_complexity_claim R S start \<equiv>
  let F = funas_trs_list (R @ S) in
  let D = defined_list (R @ S) in
  let C = filter (\<lambda>f. f \<notin> set D) F in
  case start of
    start_term.Full \<Rightarrow> Derivational_Complexity F
  | start_term.Constructor_Based \<Rightarrow> Runtime_Complexity C D"

text \<open>Desired property is decided by a given input and claim.\<close>
definition desired_property :: "('f::compare_order, 'v) input \<Rightarrow> ('f,'v) claim \<Rightarrow> bool" where
"desired_property input claim \<equiv>
  case (input,claim) of
    (Inn_TRS_input Q R S start, Terminating) \<Rightarrow>
    SN_qrel (default_nfs_trs, set (strategy_to_Q Q R), set R, set S)
  | (Inn_TRS_input Q R S start, Upperbound ub) \<Rightarrow>
    let cm = translate_complexity_claim R S start in
    let cc = Comp_Poly ub in
    deriv_bound_measure_class (rel_qrstep (True, set (strategy_to_Q Q (R @ S)), set R, set S)) cm cc
  | (CPX_input Q R S cm cc, Upperbound ub) \<Rightarrow> (* TODO: improve CPF and remove *)
    cc = Comp_Poly ub \<and>
    deriv_bound_measure_class (rel_qrstep (True, set (strategy_to_Q Q (R @ S)), set R, set S)) cm cc
  | (Inn_TRS_input Q R S start, Nonterminating) \<Rightarrow>
    \<not> SN_qrel (default_nfs_nt_trs, set (strategy_to_Q Q R), set R, set S)
  | (FP_TRS_input Outermost R, Nonterminating) \<Rightarrow> \<not> SN (ostep (lhss (set R)) (set R))
  | (FP_TRS_input Outermost R, Terminating) \<Rightarrow> SN (ostep (lhss (set R)) (set R))
  | (FP_TRS_input (Forbidden_Patterns P) R, Nonterminating) \<Rightarrow> \<not> SN (fpstep (set P) (set R))
  | (FP_TRS_input (Forbidden_Patterns P) R, Terminating) \<Rightarrow> SN (fpstep (set P) (set R))
  | (FP_TRS_input (Context_Sensitive \<mu>) R, Nonterminating) \<Rightarrow> \<not> SN (csstep (mu_of \<mu>) (set R))
  | (FP_TRS_input (Context_Sensitive \<mu>) R, Terminating) \<Rightarrow> SN (csstep (mu_of \<mu>) (set R))
  | (DP_input m P Q R, Terminating) \<Rightarrow>
    finite_dpp (default_nfs_dp, m, set P, {}, set (strategy_to_Q Q R), {}, set R)
  | (DP_input m P Q R, Nonterminating) \<Rightarrow>
    infinite_dpp (default_nfs_nt_dp, set P, set (strategy_to_Q Q R), set R) (*ignores nfs and m flags!*)
  | (Inn_TRS_input Q R S start, Confluent) \<Rightarrow> CR (rstep (set R))
  | (Inn_TRS_input Q R S start, Nonconfluent) \<Rightarrow> \<not> CR (rstep (set R))
  | (COMP_input E R, Completed) \<Rightarrow> completed_rewrite_system (set E) (set R)
  | (OCOMP_input E\<^sub>0 E R ro, Ordered_Completed) \<Rightarrow> ordered_completed_rewrite_system (set E\<^sub>0) (set E) (set R) ro
  | (EQ_input E (Equation (s,t)), Yes) \<Rightarrow> (s,t) \<in> (rstep (set E))\<^sup>\<leftrightarrow>\<^sup>*
  | (EQ_input E (Equation (s,t)), No) \<Rightarrow> (s,t) \<notin> (rstep (set E))\<^sup>\<leftrightarrow>\<^sup>*
  | (EQ_input E (Inequality (s,t)), Yes) \<Rightarrow> \<forall>\<sigma>. (s\<cdot>\<sigma>,t\<cdot>\<sigma>) \<notin> (rstep (set E))\<^sup>\<leftrightarrow>\<^sup>*
  | (EQ_input E (Inequality (s,t)), No) \<Rightarrow> \<exists>\<sigma>. (s\<cdot>\<sigma>,t\<cdot>\<sigma>) \<in> (rstep (set E))\<^sup>\<leftrightarrow>\<^sup>*
  | (CTRS_input ctrs, Terminating) \<Rightarrow> SN (cstep (set ctrs))
  | (CTRS_input ctrs, Confluent) \<Rightarrow> CR (cstep (set ctrs))
  | (CTRS_input ctrs, Nonconfluent) \<Rightarrow> \<not> (CR (cstep (set ctrs)))
  | (TA_input ta R, Yes) \<Rightarrow> (rstep (set R))^* `` ta_lang (ta_of_ta ta) \<subseteq> ta_lang (ta_of_ta ta)
  | (AC_input r a c, Terminating) \<Rightarrow> SN (relation_ac_tp (set r, set a, set c))
  | (LTS_input r, Terminating) \<Rightarrow>
    SN_on (IA.transition (lts_of r)) (IA.initial_states (lts_of r))
  | (LTS_safety_input r e, Yes) \<Rightarrow> IA.lts_safe (lts_of r) (set e)
  | (Unknown_input unk, _) \<Rightarrow> unknown_satisfied unk
  | (_, Anything) \<Rightarrow> True
  | _ \<Rightarrow> False"

end